/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/*
 * inst-sunos.C - sunos specifc code for paradynd.
 *
 * $Log: inst-sunos.C,v $
 * Revision 1.47  1998/04/22 02:30:18  buck
 * Moved showerror.h from paradynd directory to dyninstAPI directory.
 *
 * Revision 1.46  1997/07/01 16:53:14  tamches
 * added define of NULL
 *
 * Revision 1.45  1997/03/18 19:44:17  buck
 * first commit of dyninst library.  Also includes:
 * 	moving templates from paradynd to dyninstAPI
 * 	converting showError into a function (in showerror.C)
 * 	many ifdefs for BPATCH_LIBRARY in dyinstAPI/src.
 *
 * Revision 1.44  1997/02/21 20:13:37  naim
 * Moving files from paradynd to dyninstAPI + moving references to dataReqNode
 * out of the ast class. The is the first pre-dyninstAPI commit! - naim
 *
 * Revision 1.43  1997/01/16 22:04:46  tamches
 * removed flushPtrace
 *
 * Revision 1.42  1996/10/31 08:50:33  tamches
 * removed a warning
 *
 * Revision 1.41  1996/10/18 23:54:08  mjrg
 * Solaris/X86 port
 *
 * Revision 1.40  1996/08/16 21:19:03  tamches
 * updated copyright for release 1.1
 *
 * Revision 1.39  1996/08/12 16:27:12  mjrg
 * Code cleanup: removed cm5 kludges and some unused code
 *
 * Revision 1.38  1996/04/29 03:37:23  tamches
 * computePauseTimeMetric now takes in a param (but doesn't use it)
 *
 * Revision 1.37  1996/02/13 21:37:45  naim
 * Minor changes related to cost model - naim
 *
 * Revision 1.36  1996/02/12  16:43:35  naim
 * Minor change to cost model - naim
 *
 * Revision 1.35  1996/02/09  23:54:10  naim
 * Changes to the cost model - naim
 *
 * Revision 1.34  1995/12/13  23:57:31  mjrg
 * Commented out logLine message.
 *
 * Revision 1.33  1995/12/08 16:07:27  naim
 * Error message changed (adding newline) - naim
 *
 * Revision 1.32  1995/11/22  00:02:31  mjrg
 * Updates for paradyndPVM on solaris
 * Fixed problem with wrong daemon getting connection to paradyn
 * Removed -f and -t arguments to paradyn
 * Added cleanUpAndExit to clean up and exit from pvm before we exit paradynd
 * Fixed bug in my previous commit
 *
 * Revision 1.31  1995/10/19  22:36:41  mjrg
 * Added callback function for paradynd's to report change in status of application.
 * Added Exited status for applications.
 * Removed breakpoints from CM5 applications.
 * Added search for executables in a given directory.
 *
 */

#ifndef NULL
#define NULL 0
#endif

#include "dyninstAPI/src/dyninstP.h" // isApplicationPaused()
#include "dyninstAPI/src/os.h"
#include "dyninstAPI/src/dyninst.h"
#include "dyninstAPI/src/symtab.h"
#include "dyninstAPI/src/process.h"
#include "dyninstAPI/src/inst.h"
#include "dyninstAPI/src/instP.h"
#include "dyninstAPI/src/ast.h"
#include "dyninstAPI/src/ptrace_emul.h"
#include "dyninstAPI/src/util.h"
#include "dyninstAPI/src/stats.h"

#ifndef BPATCH_LIBRARY
#include "rtinst/h/trace.h"
#include "paradynd/src/metric.h"
#include "paradynd/src/main.h"
#include "paradynd/src/perfStream.h"
#include "paradynd/src/context.h"
#include "dyninstAPI/src/showerror.h"
#endif

string process::getProcessStatus() const {
   char ret[80];

   switch (status()) {
	case running:
	    sprintf(ret, "%d running", pid);
	    break;
	case neonatal:
	    sprintf(ret, "%d neonatal", pid);
	    break;
	case stopped:
	    sprintf(ret, "%d stopped", pid);
	    break;
	case exited:
	    sprintf(ret, "%d exited", pid);
	    break;
	default:
	    sprintf(ret, "%d UNKNOWN State", pid);
	    break;
    }
    return(ret);
}

//
// All costs are based on Measurements on a SPARC station 10/40.
//
void initPrimitiveCost()
{
    /* Need to add code here to collect values for other machines */

    // this doesn't really take any time
    primitiveCosts["DYNINSTbreakPoint"] = 1;

    // this happens before we start keeping time.
    primitiveCosts["DYNINSTinit"] = 1;

    primitiveCosts["DYNINSTprintCost"] = 1;

    //
    // I can't find DYNINSTincrementCounter or DYNINSTdecrementCounter
    // I think they are not being used anywhere - naim
    //
    // isthmus acutal numbers from 7/3/94 -- jkh
    // 240 ns
    primitiveCosts["DYNINSTincrementCounter"] = 16;
    // 240 ns
    primitiveCosts["DYNINSTdecrementCounter"] = 16;

#if defined(i386_unknown_solaris2_5)
    logLine("Solaris/x86 platform\n");
    // Updated calculation of the cost for the following procedures.
    // cost in cycles
    primitiveCosts["DYNINSTstartWallTimer"] = 320;
    primitiveCosts["DYNINSTstopWallTimer"] = 518;
    primitiveCosts["DYNINSTstartProcessTimer"] = 572;
    primitiveCosts["DYNINSTstopProcessTimer"] = 1143;

    // These happen async of the rest of the system.
    primitiveCosts["DYNINSTalarmExpire"] = 3724;
    primitiveCosts["DYNINSTsampleValues"] = 13;
    primitiveCosts["DYNINSTreportTimer"] = 1380;
    primitiveCosts["DYNINSTreportCounter"] = 1270;
    primitiveCosts["DYNINSTreportCost"] = 1350;
    primitiveCosts["DYNINSTreportNewTags"] = 837;

#elif defined(sparc_sun_solaris2_4)
    logLine("Solaris platform\n");
    // Updated calculation of the cost for the following procedures.
    // Clock rate = 37.04 Mhz (shemesh) - naim
    // 24.11 usecs * 37.04 Mhz  
    primitiveCosts["DYNINSTstartWallTimer"] = 893;
    // 29.08 usecs * 37.04 Mhz
    primitiveCosts["DYNINSTstopWallTimer"] = 1077;
    // 34.08 usecs * 37.04 Mhz
    primitiveCosts["DYNINSTstartProcessTimer"] = 1262;
    // 56.61 usecs * 37.04 Mhz
    primitiveCosts["DYNINSTstopProcessTimer"] = 2096;    

    // These happen async of the rest of the system.
    // 148 usecs * 37.04 Mhz
    primitiveCosts["DYNINSTalarmExpire"] = 5513;
    // 0.81 usecs * 37.04 Mhz
    primitiveCosts["DYNINSTsampleValues"] = 30;
    // 23.08 usecs * 37.04 Mhz
    primitiveCosts["DYNINSTreportTimer"] = 855;
    // 7.85 usecs * 37.04 Mhz
    primitiveCosts["DYNINSTreportCounter"] = 290;
    // 4.22 usecs * 37.04 Mhz
    primitiveCosts["DYNINSTreportCost"] = 156;
    // 1.03 usecs * 37.04 Mhz
    primitiveCosts["DYNINSTreportNewTags"] = 38;
#else
    logLine("SunOS platform\n");   
    // sparc_sun_sunos4_1_3 - default
    // The same values are used for sparc_tmc_cmost7_2 and 
    // sparc_wwt_cm5tempest1_2_3 platforms.
    // Updated calculation of the cost for the following procedures.
    // Clock rate = 67Mhz (SS-10) - naim
    if (getenv("DYNINSTuseGetrusage")) {
      // 22.08 usecs * 67Mhz
      primitiveCosts["DYNINSTstartWallTimer"] = 1479;
      // 52.76 * 67Mhz
      primitiveCosts["DYNINSTstopWallTimer"] = 3534;
      // 49.88 usecs * 67Mhz
      primitiveCosts["DYNINSTstartProcessTimer"] = 3341;
      // 69.26 usecs * 67Mhz
      primitiveCosts["DYNINSTstopProcessTimer"] = 4640;
    }
    else {
      //22.08 usecs * 67Mhz
      primitiveCosts["DYNINSTstartWallTimer"] = 1479;
      //52.76 usecs * 67Mhz
      primitiveCosts["DYNINSTstopWallTimer"] = 3534;
      //3.78 usecs * 67Mhz
      primitiveCosts["DYNINSTstartProcessTimer"] = 253;
      //6.21 usecs * 67Mhz
      primitiveCosts["DYNINSTstopProcessTimer"] = 416;
    }

    // These happen async of the rest of the system.
    // 133.86 usecs * 67Mhz
    primitiveCosts["DYNINSTalarmExpire"] = 8968;
    primitiveCosts["DYNINSTsampleValues"] = 29;
    // 6.41 usecs * 67Mhz
    primitiveCosts["DYNINSTreportTimer"] = 429;
    // 89.85 usecs * 67Mhz
    primitiveCosts["DYNINSTreportCounter"] = 6019;
    primitiveCosts["DYNINSTreportCost"] = 167;
    primitiveCosts["DYNINSTreportNewTags"] = 40; 
#endif
}

/*
 * Define the various classes of library functions to inst. 
 *
 */
void initLibraryFunctions()
{
    /* should record waiting time in read/write, but have a conflict with
     *   use of these functions by our inst code.
     *   This happens when a CPUtimer that is stopped is stopped again by the
     *   write.  It is then started again at the end of the write and should
     *   not be running then.  We could let timers go negative, but this
     *   causes a problem when inst is inserted into already running code.
     *   Not sure what the best fix is - jkh 10/4/93
     *
     */
#ifdef notdef
    tagDict["write"] = TAG_LIB_FUNC | TAG_IO_OUT;
    tagDict["read"] = TAG_LIB_FUNC | TAG_IO_IN;

    tagDict["send"] = TAG_LIB_FUNC | TAG_CPU_STATE | TAG_MSG_SEND;
    tagDict["sendmsg"] = TAG_LIB_FUNC | TAG_CPU_STATE | TAG_MSG_SEND;
    tagDict["sendto"] = TAG_LIB_FUNC | TAG_CPU_STATE | TAG_MSG_SEND;

    tagDict["rev"] = TAG_LIB_FUNC | TAG_CPU_STATE | TAG_MSG_RECV;
    tagDict["recvmsg"] = TAG_LIB_FUNC | TAG_CPU_STATE | TAG_MSG_RECV;
    tagDict["recvfrom"] = TAG_LIB_FUNC | TAG_CPU_STATE | TAG_MSG_RECV;

    tagDict["DYNINSTalarmExpire"] = TAG_LIB_FUNC;
    tagDict["DYNINSTsampleValues"] = TAG_LIB_FUNC;
    tagDict[EXIT_NAME] = TAG_LIB_FUNC;
    tagDict["fork"] = TAG_LIB_FUNC;

    tagDict["cmmd_debug"] = TAG_LIB_FUNC;
    tagDict["CMRT_init"] = TAG_LIB_FUNC;
    tagDict["CMMD_send"] = TAG_LIB_FUNC;
    tagDict["CMMD_receive"] = TAG_LIB_FUNC;
    tagDict["CMMD_receive_block"] = TAG_LIB_FUNC;
    tagDict["CMMD_send_block"] = TAG_LIB_FUNC;
    tagDict["CMMD_send_async"] = TAG_LIB_FUNC;
    tagDict["CMMD_send_async"] = TAG_LIB_FUNC;

    tagDict["main"] = 0;
#endif
}
 
#ifndef BPATCH_LIBRARY
float computePauseTimeMetric(const metricDefinitionNode *) {
    // we don't need to use the metricDefinitionNode
    timeStamp now;
    timeStamp elapsed=0.0;

    now = getCurrentTime(false);
    if (firstRecordTime) {
	elapsed = elapsedPauseTime;
	if (isApplicationPaused())
	    elapsed += now - startPause;

	assert(elapsed >= 0.0); 
	return(elapsed);
    } else {
	return(0.0);
    }
}
#endif
