/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

// shgConsts.C

/* $Log: shgConsts.C,v $
/* Revision 1.6  1997/09/24 19:23:50  tamches
/* Tk_GetFontStruct --> Tk_GetFont; XFontStruct --> Tk_Font
/* for tcl 8.0
/*
 * Revision 1.5  1996/08/16 21:07:13  tamches
 * updated copyright for release 1.1
 *
 * Revision 1.4  1996/02/15 23:10:59  tamches
 * added code to support why vs. where axis refinement
 *
 * Revision 1.3  1996/02/02 18:44:57  tamches
 * shg color change: unknown is green more more readability than tan
 *
 * Revision 1.2  1996/01/23 07:04:09  tamches
 * added shadow node features.
 * moved code here from the .h file
 *
 * Revision 1.1  1995/10/17 22:07:39  tamches
 * initial version for the new search history graph
 *
 */

#include "shgConsts.h"

shgConsts::shgConsts(Tcl_Interp *interp, Tk_Window theTkWindow) {
   display = Tk_Display(theTkWindow); // needed in destructor

   inactiveTextColor = Tk_GetColor(interp, theTkWindow, Tk_GetUid("black"));
   assert(inactiveTextColor);
   activeTextColor = Tk_GetColor(interp, theTkWindow, Tk_GetUid("ivory"));
   assert(activeTextColor);

   // Root Item FontStruct's:
   rootItemFontStruct = Tk_GetFont(interp, theTkWindow, "*-Helvetica-*-r-*-14-*");

   rootItemItalicFontStruct = Tk_GetFont(interp, theTkWindow, "*-Helvetica-*-o-*-14-*");

   // Root Item Text GCs:
   XGCValues values;
   values.foreground = activeTextColor->pixel;
   values.font = Tk_FontId(rootItemFontStruct);
   rootItemActiveTextGC = Tk_GetGC(theTkWindow, GCForeground | GCFont, &values);
   assert(rootItemActiveTextGC);

   values.foreground = inactiveTextColor->pixel;
   rootItemInactiveTextGC = Tk_GetGC(theTkWindow, GCForeground | GCFont, &values);
   assert(rootItemInactiveTextGC);

   values.font = Tk_FontId(rootItemItalicFontStruct);
   values.foreground = activeTextColor->pixel;
   rootItemActiveShadowTextGC = Tk_GetGC(theTkWindow, GCForeground | GCFont,
					 &values);
   assert(rootItemActiveShadowTextGC);

   values.foreground = inactiveTextColor->pixel;
   rootItemInactiveShadowTextGC = Tk_GetGC(theTkWindow, GCForeground | GCFont,
					   &values);
   assert(rootItemInactiveShadowTextGC);

   // Listbox FontStruct's:
   listboxItemFontStruct = Tk_GetFont(interp, theTkWindow, "*-Helvetica-*-r-*-12-*");

   listboxItemItalicFontStruct = Tk_GetFont(interp, theTkWindow, "*-Helvetica-*-o-*-12-*");

   // Listbox Item Text GCs:
   values.foreground = activeTextColor->pixel;
   values.font = Tk_FontId(listboxItemFontStruct);
   listboxItemActiveTextGC = Tk_GetGC(theTkWindow, GCForeground | GCFont, &values);
   assert(listboxItemActiveTextGC);

   values.foreground = inactiveTextColor->pixel;
   listboxItemInactiveTextGC = Tk_GetGC(theTkWindow, GCForeground | GCFont, &values);
   assert(listboxItemInactiveTextGC);

   values.font = Tk_FontId(listboxItemItalicFontStruct);
   values.foreground = activeTextColor->pixel;
   listboxItemActiveShadowTextGC = Tk_GetGC(theTkWindow, GCForeground | GCFont,
					       &values);
   assert(listboxItemActiveShadowTextGC);

   values.foreground = inactiveTextColor->pixel;
   listboxItemInactiveShadowTextGC = Tk_GetGC(theTkWindow, GCForeground | GCFont,
					      &values);
   assert(listboxItemActiveShadowTextGC);

   // 3D Borders for root item:
   rootItemTk3DBordersByStyle.resize(4);
      // indexed, in effect, by shgRootNode::evaluationState

   // Never before evaluated:
   rootItemTk3DBordersByStyle[0] = Tk_Get3DBorder(interp, theTkWindow,
						  Tk_GetUid("gray"));
   assert(rootItemTk3DBordersByStyle[0]);

   // Unknown:
   rootItemTk3DBordersByStyle[1] = Tk_Get3DBorder(interp, theTkWindow,
//						     Tk_GetUid("#e9fbb57aa3c9"));
						     Tk_GetUid("#60c0a0")); //green
   assert(rootItemTk3DBordersByStyle[1]);

   // True:
   rootItemTk3DBordersByStyle[2] = Tk_Get3DBorder(interp, theTkWindow,
						  Tk_GetUid("cornflowerblue"));
   assert(rootItemTk3DBordersByStyle[2]);

   // False:
   rootItemTk3DBordersByStyle[3] = Tk_Get3DBorder(interp, theTkWindow,
						  Tk_GetUid("pink"));
   assert(rootItemTk3DBordersByStyle[3]);

   // 3D borders for listbox:
   // It seems reasonable to use the exact same colors for shg listbox items:
   listboxItemTk3DBordersByStyle = rootItemTk3DBordersByStyle;
      // indexed, in effect, by shgRootNode::evaluationState & an active flag

   // Refinement Styles (why axis, where axis)
   whyRefinementColor = Tk_GetColor(interp, theTkWindow, Tk_GetUid("palegoldenrod"));
   whereRefinementColor = Tk_GetColor(interp, theTkWindow, Tk_GetUid("orchid"));

   values.foreground = whyRefinementColor->pixel;
   values.line_width = 2;
   whyRefinementRayGC = Tk_GetGC(theTkWindow, GCForeground | GCLineWidth, &values);
   assert(whyRefinementRayGC);

   values.foreground = whereRefinementColor->pixel;
   values.line_width = 2;
   whereRefinementRayGC = Tk_GetGC(theTkWindow, GCForeground | GCLineWidth, &values);
   assert(whereRefinementRayGC);
}

shgConsts::~shgConsts() {
   Tk_FreeFont(rootItemFontStruct);
   Tk_FreeFont(rootItemItalicFontStruct);
   Tk_FreeFont(listboxItemFontStruct);
   Tk_FreeFont(listboxItemItalicFontStruct);
 
   Tk_FreeColor(activeTextColor);
   Tk_FreeColor(inactiveTextColor);

   Tk_FreeGC(display, rootItemActiveTextGC);
   Tk_FreeGC(display, rootItemInactiveTextGC);
   Tk_FreeGC(display, rootItemActiveShadowTextGC);
   Tk_FreeGC(display, rootItemInactiveShadowTextGC);
   Tk_FreeGC(display, listboxItemActiveTextGC);
   Tk_FreeGC(display, listboxItemInactiveTextGC);
   Tk_FreeGC(display, listboxItemActiveShadowTextGC);
   Tk_FreeGC(display, listboxItemInactiveShadowTextGC);

   Tk_FreeGC(display, whyRefinementRayGC);
   Tk_FreeGC(display, whereRefinementRayGC);

   Tk_FreeColor(whyRefinementColor);
   Tk_FreeColor(whereRefinementColor);

   for (unsigned stylelcv=0; stylelcv < rootItemTk3DBordersByStyle.size(); stylelcv++)
      Tk_Free3DBorder(rootItemTk3DBordersByStyle[stylelcv]);

   // Note that we intentionally don't free up anything in
   // listboxItemTk3DBordersByStyle, since it was always just a shadow copy
   // of the contents of rootItemTk3DBordersByStyle
}
